unit UnitGMaps;

interface
uses windows,types,graphics,jpeg;

const ZoomMax=20;

type
  TGeoPoint=packed record
  case integer of
    0:(Lon,Lat:double);
    1:(x,y:double);
  end;

  TRegionGlobe=packed record
  case integer of
    0:(LonMin,LatMin,LonMax,LatMax:double);
    1:(Min,Max:TGeopOint);
  end;

  TConstantes=record
    pixelsPerLonDegree:double;
	  pixelsPerLonRadian:double;
	  numTiles:integer;
	  bitmapOrigo:TGeoPoint;
  end;

procedure LoadToBitMap(ABitMap:TBitMap;BBox:TRegionGlobe);

implementation
uses math,classes,IdHTTP,sysutils;
const tileSize=256;


function LevelTileSizeToDegrees(TileL:cardinal):double;
// degrees for level 0
const LevelZeroTileSizeDegrees =5.625;  // 360 / 64 = 360 / (2 ^ 6)
begin
  result:= 360 /(1 shl TileL);
end;

function fmod(x,y:double):double;
var e:integer;
begin
  e:=trunc(x/y);
  result:=x-y*e;
end;

function getKeyholeDirection(x,y:cardinal):string;
begin
  result:='';
  if x=1
  then begin
         if y=1 then result:='s'
	              else if y=0 then result:='r';
       end
  else
  if x=0
  then begin
          if y=1 then	result:='t'
          else if y=0then	result:='q';
       end;
end;

function getKeyholeString(x,y,z:cardinal):string;
var s:string;
    i:integer;
    xx,yy:double;
    rx,ry:integer;
begin
	s:='';
  xx:=x;
  yy:=y;

  i:=17;
	while i>=z do
  begin
		rx:=round(fmod(xx,2));
		xx:=floor(xx/2);
		ry:=round(fmod(yy,2));
		yy:=floor(yy/2);
		s:=getKeyholeDirection(rx,ry)+s;
    dec(i);
  end;
	result:='t'+s;
end;



function GetConstants(z:cardinal):TConstantes;
var CentreX:double;
    NumPixels:integer;
begin
  with result do
  begin
    numTiles:=(1 shl (zoomMax-z));
    NumPixels:=tileSize*numTiles;
  	pixelsPerLonDegree:=NumPixels/360;
    pixelsPerLonRadian:=NumPixels/(2*Pi);
    CentreX:=NumPixels/2;
    bitmapOrigo.x:=CentreX;
    bitmapOrigo.y:=CentreX;
  end;
end;

function getBitmapCoordinate(Lat,Lon:double;zoomLevel:cardinal):TGeoPoint;
var d:TGeoPoint;
    Constants:TConstantes;
    e:double;
begin
  d.x:=0;
  d.y:=0;
  Constants:=GetConstants(zoomLevel);
  d.x:=floor(Constants.bitmapOrigo.x + Lon*Constants.pixelsPerLonDegree);
  e:=sin(Lat*PI/180);

  if e>0.9999 then e:=0.9999;
  if e<-0.9999 then e:=-0.9999;

  d.y:=floor(Constants.bitmapOrigo.y-0.5*ln((1+e)/(1-e))*Constants.pixelsPerLonRadian);

  Result:=d;
end;

function GetTileXY(Lat,Lon:double;zoomLevel:cardinal):TPoint;
var d:TGeoPoint;
begin
   d:=getBitmapCoordinate(Lat, Lon, zoomLevel);
   result.x:=floor(d.x/tileSize);
   result.y:=floor(d.y/tileSize);
end;

function GetTileBox(X,Y,zoom:cardinal):TRegionGlobe;
var d:TGeoPoint;
    C:TConstantes;
    Lon,Lat,e,f,g:double;
begin
  d.x:=x*tileSize;
  d.y:=y*tileSize;
  C:=GetConstants(zoom);
  Lon:=(d.x-c.bitmapOrigo.x)/c.pixelsPerLonDegree;
  f:=2*(c.bitmapOrigo.y-d.y)/c.pixelsPerLonRadian;
  g:=exp(f);
  e:=(g-1)/(g+1);
  Lat:=ArcSin(e)*180/Pi;
  Result.LonMin:=Lon;
  Result.LatMax:=Lat;
  d.x:=(x+1)*tileSize;
  d.y:=(y+1)*tileSize;
  C:=GetConstants(zoom);
  Lon:=(d.x-c.bitmapOrigo.x)/c.pixelsPerLonDegree;
  f:=2*(c.bitmapOrigo.y-d.y)/c.pixelsPerLonRadian;
  g:=exp(f);
  e:=(g-1)/(g+1);
  Lat:=ArcSin(e)*180/Pi;
  Result.LonMax:=Lon;
  Result.LatMin:=Lat;
end;

function getTileLatLong(Lat,Lon:double;zoomLevel:cardinal):TGeoPoint;
var d,e:TGeoPoint;
    a,b,f:double;
    c:TConstantes;
begin
 d.x:=0;
 d.y:=0;
 c:=GetConstants(zoomLevel);

 e:=getBitmapCoordinate(Lat, Lon, zoomLevel);
 a:=e.x;
 b:=e.y;

 d.x:=(a-c.bitmapOrigo.x)/c.pixelsPerLonDegree;
 f:=-(b-c.bitmapOrigo.y)/c.pixelsPerLonRadian;
 d.y:=360*(ArcTan2(exp(f),1)-PI/2)/pi;

 result:=d;
end;

function BBoxToTileBox(BBox:TRegionGlobe;zoomLevel:cardinal):TRect;
begin
  Result.TopLeft:=GetTileXY(BBox.LatMax,bbox.LonMin,zoomLevel);
  Result.BottomRight:=GetTileXY(BBox.LatMin,bbox.LonMax,zoomLevel);
end;

function getSatURL( TileL, TileX, TileY: cardinal) : string;
var
 gx, gy, gl: double;
 wx, wy, wl: double;
 tx, ty: double;
 LevelTileSizeDegrees: double;
 tl:integer;
 tid: string;
 tbox:TRegionGlobe;
 bboxtile:TRect;
 bboxtiledim:cardinal;
 quadrant:smallint;
 x,y,x2,y2:cardinal;
 c:char;
begin
 tid := 't';
 bboxtiledim:=(1 shl TileL);
 bboxtile:=rect(1,1,bboxtiledim,bboxtiledim);
 x:=TileX+1;
 y:=tileY+1;
 while bboxtiledim>1 do
 begin
   x2:=bboxtile.Left+bboxtiledim div 2-1;
   y2:=bboxtile.Top+bboxtiledim div 2-1;
   quadrant:=0;
   if x>x2 then quadrant:=1;
   if y>y2 then quadrant:=quadrant+2;

   case quadrant of
     0:begin
         bboxtile.Right:=x2;
         bboxtile.Bottom:=y2;
         c:='q';
       end;
     1:begin
         bboxtile.Left:=x2+1;
         bboxtile.Bottom:=y2;
         c:='r';
       end;
     2:begin
         bboxtile.Right:=x2;
         bboxtile.Top:=y2+1;
         c:='t';
       end;
     3:begin
         bboxtile.Left:=x2+1;
         bboxtile.Top:=y2+1;
         c:='s';
       end;
   end;
   tid:=tid+c;
   bboxtiledim:=bboxtiledim div 2;
 end;

 result := 'http://kh0.google.com/kh?n=404&v=9&t=' + tid;

end;

function LoadTile(ATile:TJPEGImage;X,Y,zoom:cardinal):boolean;
var AIdHTTP:TIdHTTP;
    MS:TMemoryStream;
    AURL:string;
begin
  result:=false;
  AURL:=getSatURL(zoomMax-zoom,X,Y);
  AIdHTTP:=TIdHTTP.Create(nil);

//  AIdHTTP.ProxyParams.ProxyPort:=XXXX;
//  AIdHTTP.ProxyParams.ProxyServer:='XXX.XX.XXX.XX';
//  AIdHTTP.ProxyParams.BasicAuthentication:=true;
//  AIdHTTP.ProxyParams.ProxyUsername:='XXXXXXX';
//  AIdHTTP.ProxyParams.ProxyPassword:='XXXXX';



  MS:=TMemoryStream.Create;
  try
    try
    AIdHTTP.Get(AURL,MS);
    except
    end;
    if AIdHTTP.ResponseCode=200 then
    begin
      MS.Seek(0, soFromBeginning);
      ATile.LoadFromStream(MS);
      //ATile.SaveToFile('T'+IntToStr(X)+'_'+IntToStr(Y)+'.jpeg');
      result:=true;
    end;
  finally
    AIdHTTP.Free;
    MS.Free;
  end;
end;

procedure LoadToBitMap0(ABitMap:TBitMap;ARect:TRect;BBox:TRegionGlobe;Zoom:integer);
var AJPeg:TJPEGImage;
    i,j:integer;
    RectDest,ClipSourceRect,SourceRect:trect;
    c:TConstantes;
    bboxtile:TRegionGlobe;
    degW,degH:double;
    rW,rH,a,b:double;
    bmp,bmp2:tbitmap;
begin
  c:=GetConstants(zoom);

  degW:=(BBox.LonMax-BBox.LonMin);
  degH:=(BBox.LatMax-BBox.LatMin);
  rH:=ABitMap.Height/degH;
  rW:=ABitMap.Width/degW;

  AJPeg:=TJPEGImage.Create;
  bmp:=TBitmap.Create;
  bmp2:=TBitmap.Create;
  try
  for i:=ARect.Left to ARect.Right do
    for j:=ARect.Top to ARect.Bottom do
    begin
      bboxtile:=GetTileBox(i,j,zoom);
      a:=rW;
      b:=-a*BBox.LonMin;
      rectdest.Left:=floor(a*bboxtile.LonMin+b);
      RectDest.Right:=ceil(a*bboxtile.LonMax+b);

      a:=-rH;
      b:=-a*BBox.LatMax;
      rectdest.top:=floor(a*bboxtile.LatMax+b);
      rectdest.bottom:=ceil(a*bboxtile.LatMin+b);

      SourceRect:=rect(0,0,256,256);
      a:=256/(rectdest.right-rectdest.left);
      b:=-a*rectdest.Left;
      if rectdest.Left<0 then sourceRect.Left:=floor(b);
      if rectdest.right>ABitMap.Width then sourceRect.right:=ceil(a*ABitMap.Width+b);

      a:=256/(rectdest.bottom-rectdest.top);
      b:=-a*rectdest.top;
      if rectdest.top<0 then sourceRect.top:=floor(b);
      if rectdest.bottom>ABitMap.Height then sourceRect.bottom:=ceil(a*ABitMap.Height+b);

      with RectDest do
      begin
        if Left<0 then Left:=0;
        if Right>ABitMap.Width then Right:=ABitMap.Width;
        if Top<0 then Top:=0;
        if Bottom>ABitMap.Height then Bottom:=ABitMap.Height;
      end;

      if LoadTile(AJPeg,i,j,zoom)
      then begin
             bmp.Width:=SourceRect.Right-SourceRect.Left;
             bmp.Height:=SourceRect.bottom-SourceRect.top;
             bmp.Canvas.Draw(-SourceRect.Left,-SourceRect.Top,AJPeg);
             //bmp.Assign(ajpeg);
//             bmp.Canvas.FrameRect(sourcerect);
             //bmp.SaveToFile('t_'+IntToStr(i)+'_'+IntToStr(j)+'.bmp');
             sleep(20);
//             InflateRect(rectdest,1,1);
//             ABitMap.Canvas.CopyRect(RectDest,bmp.Canvas,SourceRect);
             //ABitMap.Canvas.Brush.Color:=clblack;
//             ABitMap.Canvas.FrameRect(RectDest);
             ABitMap.Canvas.StretchDraw(RectDest,bmp);
           end;
//      else
    end;
  finally
    AJPeg.Free;
    bmp.Free;
    bmp2.Free;
  end;
end;

procedure LoadToBitMap(ABitMap:TBitMap;BBox:TRegionGlobe);
var ARect:TRect;
    zoom:integer;
    pixelsPerLonDegreeMini:double;
    AspectRatio:double;// (=rapport largeur/hauteur)
    c:TConstantes;
begin
  AspectRatio:=(BBox.LonMax-BBox.LonMin)/(BBox.LatMax-BBox.LatMin)*3/4;
  if AspectRatio>1
  then ABitMap.height:=round(ABitMap.width/AspectRatio)
  else ABitMap.width:=round(ABitMap.height*AspectRatio);


  pixelsPerLonDegreeMini:=ABitMap.Width/(BBox.LonMax-BBox.LonMin);

  zoom:=zoommax;
  c:=GetConstants(zoom);
  while ((zoom>0) and (c.pixelsPerLonDegree<pixelsPerLonDegreeMini)) do
  begin
    dec(zoom);
    c:=GetConstants(zoom);
  end;

  AREct:=BBoxToTileBox(BBox,zoom);
  LoadToBitMap0(ABitMap,ARect,bbox,zoom);
end;

end.
